//
//  ST.CPP
//
//  Michael Schindler
//  September 25, 1996
//  http://eiunix.tuwien.ac.at/~michael
//
// DESCRIPTION
// -----------
//
//  This program performs a S transform on an input file or stream,
//  and sends the result to an output file or stream.
//
//  This program takes two arguments: an input file and an output
//  file.  You can leave off one argument and send your output to
//  stdout.  Leave off two arguments and read your input from stdin
//  as well.
//
//  The ST output looks like this:
//
//  long blocksize   { start-char  ...data... }*
//
//  This program accompanies Mark Nelsons programs to his article
//  "Data Compression with the Burrows-Wheeler Transform."
//
//
// Build Instructions
// ------------------
//
//  Define the constant unix for UNIX or UNIX-like systems.  The
//  use of this constant turns off the code used to force the MS-DOS
//  file system into binary mode.  g++ already does this, your UNIX
//  C++ compiler might also.
//
//  Borland C++ 4.5 16 bit    : bcc -w -ml st.cpp
//  Borland C++ 4.5 32 bit    : bcc32 -w st.cpp
//  Microsoft Visual C++ 1.5  : cl /W4 /AL st.cpp
//  Microsoft Visual C++ 4.0  : cl /W4 st.cpp
//  g++                       : g++ -o st st.cpp
//
// Typical Use
// -----------
//
//  st < raw-file | mtf | rle | ari > compressed-file
//

#include <stdio.h>
#if !defined( unix )
#include <io.h>
#endif
#include <string.h>

#if ( INT_MAX == 32767 )
#define SIZE_SHIFT 14
#else
#define SIZE_SHIFT 18
#endif
#define BLOCK_SIZE (1 << SIZE_SHIFT)

unsigned char inbuffer[ BLOCK_SIZE ];
unsigned char outbuffer[ BLOCK_SIZE ];
unsigned int buflen;
unsigned int count[ 256 ];

main( int argc, char *argv[] )
{
    int i, sum;
    fprintf( stderr, "Performing ST on " );
    if ( argc > 1 ) {
        freopen( argv[ 1 ], "rb", stdin );
        fprintf( stderr, "%s", argv[ 1 ] );
    } else
        fprintf( stderr, "stdin" );
    fprintf( stderr, " to " );
    if ( argc > 2 ) {
        freopen( argv[ 2 ], "wb", stdout );
        fprintf( stderr, "%s", argv[ 2 ] );
    } else
        fprintf( stderr, "stdout" );
    fprintf( stderr, "\n" );
#if !defined( unix )
    setmode( fileno( stdin ), O_BINARY );
    setmode( fileno( stdout ), O_BINARY );
#endif
    buflen = BLOCK_SIZE;
    putc( (char) SIZE_SHIFT, stdout );
    for ( ; ; ) {
        buflen = fread( (char *)inbuffer, 1, (size_t)BLOCK_SIZE, stdin);
        if (buflen == 0) break;
        fprintf( stderr,
                 "Processing %ld bytes\n",
                 buflen );
        for (i=0; i<256; i++)
            count[ i ] = 0;                             // set counters to zero
        for (i=0; i<buflen; i++)
            count[ inbuffer[ i ] ]++;                   // count frequencies
        sum = 0;
        for (i=0; i<256; i++) {
            sum += count[i];
            count[i] = sum - count[i];
        }
        for ( i=1; i<buflen; i++ )
             outbuffer[ count[inbuffer[i-1]]++ ] = inbuffer[ i ];
        outbuffer[ count[inbuffer[buflen-1]]++ ] = inbuffer[ 0 ];

        putc( inbuffer[ 0 ], stdout );
        fwrite( (char *)outbuffer, 1, (size_t)buflen, stdout );
    }
    return 1;
}

